#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
import time
from typing import Callable, List, TypeVar

from .exceptions import RetryableFetchError

T = TypeVar("T")


def retry(f: Callable[[], T], backoff_times: List[int] = None, retry_number=1) -> T:
    """
    Calls f() and returns its result. However, if f() throws an exception, we will retry.
    Each entry in backoff_times is a length of time (in seconds) to sleep using `time.sleep`.
    So if the backoff_times = [1, 4, 9], we will sleep 1 second, then 4 seconds, then 9 seconds between retries.

    Also, if an exception should be raised on the last attempt, we will attempt to add a "retries" attribute to
    it with the number of retries that were attempted.

    :param f: the function to call
    :param backoff_times: a list of times to wait
    :param retry_number: the current "retry" (used only for the retries attribute)
    :return: whatever the result of `f` is
    """
    exception = None
    try:
        return f()
    except RetryableFetchError as ex:
        if backoff_times:
            # if we have, say, [1, 4, 9] in our list of backoff times,
            # we will delay for 1 second and then have [4, 9] remaining backoff times
            delay, *remaining_backoff_times = backoff_times

            # sleep
            time.sleep(delay)

            return retry(f, remaining_backoff_times, retry_number + 1)
        else:
            # if this exception has a slot for "retry count" we should set it
            if hasattr(ex, "retries"):
                ex.retries = retry_number

            # now raise the exception
            raise ex
